"""
Integration tests with conda
"""

import json
import os
import sys
from contextlib import contextmanager
from pathlib import Path
from subprocess import check_output
from tempfile import NamedTemporaryFile

import pytest
from conda.models.version import VersionOrder
from conda.testing.integration import run_command
from conftest import BASE_PREFIX, DATA, PLATFORM

from menuinst._schema import validate
from menuinst.platforms import Menu, MenuItem

ENV_VARS = {
    k: v
    for (k, v) in os.environ.copy().items()
    if not k.startswith(("CONDA", "_CONDA", "MAMBA", "_MAMBA"))
}
ENV_VARS["CONDA_VERBOSITY"] = "3"


@contextmanager
def new_environment(tmpdir, *packages):
    try:
        prefix = str(tmpdir / "prefix")
        print("--- CREATING", prefix, "---")
        stdout, stderr, retcode = run_command(
            "create", prefix, "-y", "--offline", *[str(p) for p in packages]
        )
        assert not retcode
        for stream in (stdout, stderr):
            if "menuinst Exception" in stream:
                raise RuntimeError(
                    f"Creation command exited with 0 but stdout contained exception:\n{stream}"
                )

        yield prefix
    finally:
        print("--- REMOVING", prefix, "---")
        stdout, stderr, retcode = run_command("remove", prefix, "--offline", "--all")
        assert not retcode
        for stream in (stdout, stderr):
            if "menuinst Exception" in stream:
                raise RuntimeError(
                    f"Deletion Command exited with 0 but stdout contained exception:\n{stream}"
                )


@contextmanager
def install_package_1(tmpdir):
    """
    This package is shipped with the test data and contains two menu items.

    Both will echo the `CONDA_PREFIX` environment variable. However, the
    first one preactivates the environment, while the second does not. This
    means that the first shortcut will successfully echo the prefix path,
    while the second one will be empty (Windows) or "N/A" (Unix).
    """
    with new_environment(tmpdir, DATA / "pkgs" / "noarch" / "package_1-0.1-0.tar.bz2") as prefix:
        menu_file = Path(prefix) / "Menu" / "package_1.json"
        with open(menu_file) as f:
            meta = json.load(f)
            assert len(meta["menu_items"]) == 2
        assert menu_file.is_file()
        yield prefix, menu_file
    assert not menu_file.is_file()


def test_conda_recent_enough():
    data = json.loads(check_output([sys.executable, "-I", "-m", "conda", "info", "--json"]))
    assert VersionOrder(data["conda_version"]) >= VersionOrder("4.12a0")


@pytest.mark.skipif(PLATFORM != "linux", reason="Linux only")
def test_package_1_linux(tmpdir):
    applications_menu = Path(tmpdir) / "config" / "menus" / "applications.menu"
    if applications_menu.is_file():
        original_xml = applications_menu.read_text()
    else:
        original_xml = None
    with install_package_1(tmpdir) as (prefix, menu_file):
        meta = validate(menu_file)
        menu = Menu(meta.menu_name, str(prefix), BASE_PREFIX)
        menu_items = [item.dict() for item in meta.menu_items]
        items = [menu]

        # First case, activation is on, output should be the prefix path
        # Second case, activation is off, output should be N/A
        for item, expected_output in zip(menu_items, (str(prefix), "N/A")):
            item = MenuItem(menu, item)
            items.append(item)
            command = item._command()
            print(command)
            print("-----")
            output = check_output(command, shell=True, universal_newlines=True, env=ENV_VARS)
            assert output.strip() == expected_output

    assert not Path(prefix).exists()
    for item in items:
        for path in item._paths():
            assert not path.exists()

    if original_xml:
        assert original_xml == applications_menu.read_text()


@pytest.mark.skipif(PLATFORM != "osx", reason="MacOS only")
def test_package_1_osx(tmpdir):
    with install_package_1(tmpdir) as (prefix, menu_file):
        meta = validate(menu_file)
        menu_items = [item.dict() for item in meta.menu_items]
        menu = Menu(meta.menu_name, str(prefix), BASE_PREFIX)
        items = [menu]
        # First case, activation is on, output should be the prefix path
        # Second case, activation is off, output should be N/A
        for item, expected_output in zip(menu_items, (str(prefix), "N/A")):
            item = MenuItem(menu, item)
            items.append(item)
            script = item._write_script(
                script_path=NamedTemporaryFile(suffix=".sh", delete=False).name
            )
            print(item._command())
            print("-------------")
            output = check_output(["bash", script], universal_newlines=True, env=ENV_VARS)
            Path(script).unlink()
            assert output.strip() == expected_output

    assert not Path(prefix).exists()
    for item in items:
        for path in item._paths():
            assert not path.exists()


@pytest.mark.skipif(PLATFORM != "win", reason="Windows only")
def test_package_1_windows(tmpdir):
    with install_package_1(tmpdir) as (prefix, menu_file):
        meta = validate(menu_file)
        menu = Menu(meta.menu_name, str(prefix), BASE_PREFIX)
        menu_items = [item.dict() for item in meta.menu_items]
        items = [menu]
        # First case, activation is on, output should be the prefix path
        # Second case, activation is off, output should be empty
        for item, expected_output in zip(menu_items, (str(prefix), "!CONDA_PREFIX!")):
            item = MenuItem(menu, item)
            items.append(item)
            script = item._write_script(
                script_path=NamedTemporaryFile(suffix=".bat", delete=False).name
            )
            print(item._command())
            print("-------------")
            output = check_output(
                ["cmd.exe", "/C", f"conda deactivate && conda deactivate && {script}"],
                universal_newlines=True,
                env=ENV_VARS,
            )
            Path(script).unlink()
            output = output.replace("ECHO is off.", "")
            assert output.splitlines()[0].strip() == expected_output

    assert not Path(prefix).exists()
    for item in items:
        for path in item._paths():
            assert not path.exists()



"""
"""

import os
from logging import basicConfig, getLogger

from .api import _install_adapter as install

try:
    from ._version import __version__
except ImportError:
    __version__ = "dev"

log = getLogger(__name__)
if os.environ.get("MENUINST_DEBUG"):
    basicConfig(level="DEBUG")

__all__ = ["install", "__version__"]


# Compatibility forwarders for menuinst v1.x (Windows only)
if os.name == "nt":
    from ._vendor.apipkg import initpkg

    initpkg(
        __name__,
        exportdefs={
            "win32": {
                "dirs_src": "menuinst.platforms.win_utils.knownfolders:dirs_src",
            },
            "knownfolders": {
                "get_folder_path": "menuinst.platforms.win_utils.knownfolders:get_folder_path",
                "FOLDERID": "menuinst.platforms.win_utils.knownfolders:FOLDERID",
            },
            "winshortcut": {
                "create_shortcut": "menuinst.platforms.win_utils.winshortcut:create_shortcut",
            },
            "win_elevate": {
                "runAsAdmin": "menuinst.platforms.win_utils.win_elevate:runAsAdmin",
                "isUserAdmin": "menuinst.platforms.win_utils.win_elevate:isUserAdmin",
            },
        },
        # Calling initpkg WILL CLEAR the 'menuinst' top-level namespace, and only then will add
        # the exportdefs contents! If we want to keep something defined in this module, we MUST
        # make sure it's added in the 'attr' dictionary below.
        attr={"__version__": __version__, "install": install},
    )



import pytest
from conftest import DATA

# from hypothesis import given, settings, HealthCheck
# from hypothesis_jsonschema import from_schema
from pydantic import ValidationError as ValidationErrorV2
from pydantic.v1 import ValidationError as ValidationErrorV1

from menuinst._schema import BasePlatformSpecific, MenuItem, validate

# # suppress_health_check=3 --> too_slow
# @settings(max_examples=100, suppress_health_check=[HealthCheck.too_slow])
# @given(from_schema(MenuInstSchema.schema()))
# def test_schema_can_be_loaded(value):
#     assert value


@pytest.mark.parametrize(
    "path", [pytest.param(path, id=path.name) for path in sorted((DATA / "jsons").glob("*.json"))]
)
def test_examples(path):
    if "invalid" in path.name:
        with pytest.raises((ValidationErrorV1, ValidationErrorV2)):
            assert validate(path)
    else:
        assert validate(path)


def test_MenuItemMetadata_synced_with_OptionalMenuItemMetadata():
    fields_as_required = MenuItem.__fields__.copy()
    fields_as_required.pop("platforms", None)
    fields_as_optional = BasePlatformSpecific.__fields__
    assert fields_as_required.keys() == fields_as_optional.keys()
    for (_, required), (_, optional) in zip(
        sorted(fields_as_required.items()), sorted(fields_as_optional.items())
    ):
        assert required.field_info.description == optional.field_info.description



import os
import sys

import pytest
from conda.cli.python_api import run_command

if sys.platform == "win32":
    import menuinst._legacy as menuinst
    from menuinst._legacy.win32 import dirs_src


def file_exist(mode, name):
    file = os.path.join(
        dirs_src[mode]['start'][0], 'Anaconda3 (64-bit) - Test Menu', '%s.lnk' % name
    )
    return os.path.exists(file)


menu_dir = os.path.dirname(__file__)


@pytest.mark.skipif(sys.platform != "win32", reason="Windows-only tests")
class TestWindowsShortcuts(object):
    def test_install_folders_exist(self):
        for mode in ["user", "system"]:
            for path, _ in dirs_src[mode].values():
                assert os.path.exists(path)

    def test_create_and_remove_shortcut(self):
        nonadmin = os.path.join(sys.prefix, ".nonadmin")
        shortcut = os.path.join(menu_dir, "menu-windows.json")
        has_nonadmin = os.path.exists(nonadmin)
        name = 'Anaconda Prompt'
        for mode in ["user", "system"]:
            if mode == "user":
                open(nonadmin, 'a').close()
            menuinst.install(shortcut, remove=False)
            assert file_exist(mode, name)
            menuinst.install(shortcut, remove=True)
            assert not file_exist(mode, name)
            if os.path.exists(nonadmin):
                os.remove(nonadmin)
        if has_nonadmin:
            open(nonadmin, 'a').close()

    def test_create_shortcut_env(self):
        nonadmin = os.path.join(sys.prefix, ".nonadmin")
        open(nonadmin, 'a').close()
        shortcut = os.path.join(menu_dir, "menu-windows.json")
        test_env_name = 'test_env'
        run_command("create", "-n", test_env_name, "python")
        prefix = os.path.join(sys.prefix, 'envs', test_env_name)
        name = 'Anaconda Prompt (%s)' % test_env_name
        menuinst.install(shortcut, prefix=prefix, remove=False)
        assert file_exist('user', name)
        menuinst.install(shortcut, prefix=prefix, remove=True)
        assert not file_exist('user', name)
        run_command("remove", "-n", test_env_name, "--all")

    def test_root_prefix(self):
        nonadmin = os.path.join(sys.prefix, ".nonadmin")
        open(nonadmin, 'a').close()
        shortcut = os.path.join(menu_dir, "menu-windows.json")
        root_prefix = os.path.join(menu_dir, 'temp_env')
        run_command("create", "-p", root_prefix, "python")
        name = 'Anaconda Prompt (%s)' % os.path.split(sys.prefix)[1]
        menuinst.install(shortcut, remove=False, root_prefix=root_prefix)
        assert file_exist('user', name)
        menuinst.install(shortcut, remove=True, root_prefix=root_prefix)
        assert not file_exist('user', name)
        run_command("remove", "-p", root_prefix, "--all")



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import sys
from logging import getLogger

import pytest

if sys.platform == "win32":
    from menuinst._legacy.win32 import quote_args

log = getLogger(__name__)


@pytest.mark.skipif(sys.platform != "win32", reason="Windows-only tests")
def test_quote_args_1():
    args = [
        "%windir%\\System32\\cmd.exe",
        "/K",
        "c:\\Users\\Francisco García Carrión Martínez\\Anaconda 3\\Scripts\\activate.bat",
        "c:\\Users\\Francisco García Carrión Martínez\\Anaconda 3",
    ]
    assert quote_args(args) == [
        "\"%windir%\\System32\\cmd.exe\"",
        "/K",
        "\"\"c:\\Users\\Francisco García Carrión Martínez\\Anaconda 3\\Scripts\\activate.bat\" \"c:\\Users\\Francisco García Carrión Martínez\\Anaconda 3\"\"",  # noqa
    ]



"""Ensure JSON schemas are up-to-date with code"""

import json

from menuinst._schema import dump_default_to_json, dump_schema_to_json
from menuinst.utils import data_path


def test_schema_is_up_to_date():
    with open(data_path("menuinst.schema.json")) as f:
        in_file = json.load(f)
    in_code = dump_schema_to_json(write=False)
    assert in_file == in_code


def test_defaults_are_up_to_date():
    with open(data_path("menuinst.default.json")) as f:
        in_file = json.load(f)
    in_code = dump_default_to_json(write=False)
    assert in_file == in_code



import json
import logging
import os
import shutil
from pathlib import Path
from subprocess import check_output
from tempfile import TemporaryDirectory

import py
import pytest

from menuinst.platforms.base import platform_key

logging.basicConfig(level=logging.DEBUG)

os.environ["PYTEST_IN_USE"] = "1"
DATA = Path(__file__).parent / "data"
PLATFORM = platform_key()


def base_prefix():
    prefix = os.environ.get("CONDA_ROOT", os.environ.get("MAMBA_ROOT_PREFIX"))
    if not prefix:
        prefix = json.loads(check_output(["conda", "info", "--json"]))["root_prefix"]
    return prefix


BASE_PREFIX = base_prefix()


@pytest.fixture()
def delete_files():
    paths = []
    yield paths
    for path in paths:
        path = Path(path)
        # If the list contains duplicates, a path may already have been deleted.
        if not path.exists():
            continue
        try:
            if path.is_dir():
                shutil.rmtree(path)
            else:
                path.unlink()
        except IOError:
            logging.warning("Could not delete %s", path, exc_info=True)


@pytest.fixture(scope="function")
def tmpdir(tmpdir, request):
    Path(str(tmpdir)).mkdir(parents=True, exist_ok=True)
    tmpdir = TemporaryDirectory(dir=str(tmpdir))
    request.addfinalizer(tmpdir.cleanup)
    return py.path.local(tmpdir.name)


@pytest.fixture(autouse=True)
def mock_locations(monkeypatch, tmp_path):
    from menuinst.platforms.linux import LinuxMenu
    from menuinst.platforms.osx import MacOSMenuItem

    if os.name == "nt":
        from menuinst.platforms.win_utils import knownfolders

        def windows_locations(preferred_mode, check_other_mode, key):
            return tmp_path / key

        monkeypatch.setattr(knownfolders, "folder_path", windows_locations)

    def osx_base_location(self):
        return tmp_path

    if not os.environ.get("CI"):
        monkeypatch.setattr(MacOSMenuItem, "_base_location", osx_base_location)

    # For Linux
    if not os.environ.get("CI"):
        monkeypatch.setattr(LinuxMenu, "_system_config_directory", tmp_path / "config")
        monkeypatch.setattr(LinuxMenu, "_system_data_directory", tmp_path / "data")
        monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path / "config"))
        monkeypatch.setenv("XDG_DATA_HOME", str(tmp_path / "data"))


@pytest.fixture()
def run_as_user(monkeypatch):
    from menuinst import utils as menuinst_utils

    monkeypatch.setattr(menuinst_utils, "user_is_admin", lambda: False)



""""""

import json
import os
import plistlib
import shlex
import subprocess
import sys
import warnings
from pathlib import Path
from tempfile import NamedTemporaryFile, mkdtemp
from time import sleep, time
from typing import Iterable, Tuple
from xml.etree import ElementTree

import pytest
from conftest import DATA, PLATFORM

from menuinst.api import install, remove
from menuinst.platforms.osx import _lsregister
from menuinst.utils import DEFAULT_PREFIX, logged_run, slugify


def _poll_for_file_contents(path, timeout=10):
    t0 = time()
    while not os.path.isfile(path):
        sleep(1)
        if time() >= t0 + timeout / 2:
            raise RuntimeError(f"Timeout. File '{path}' was not created!")

    out = ""
    while not (out := Path(path).read_text()):
        sleep(1)
        if time() >= t0 + timeout:
            raise RuntimeError(f"Timeout. File '{path}' was empty!")
    return out


def check_output_from_shortcut(
    delete_files,
    json_path,
    remove_after=True,
    expected_output=None,
    action="run_shortcut",
    file_to_open=None,
    url_to_open=None,
) -> Tuple[Path, Iterable[Path], str]:
    assert action in ("run_shortcut", "open_file", "open_url", None)

    output = None
    output_file = None
    abs_json_path = DATA / "jsons" / json_path
    contents = abs_json_path.read_text()
    if "__OUTPUT_FILE__" in contents:
        with NamedTemporaryFile(suffix=json_path, mode="w", delete=False) as tmp:
            output_file = str(Path(tmp.name).resolve()) + ".out"
            contents = contents.replace("__OUTPUT_FILE__", output_file.replace("\\", "\\\\"))
            tmp.write(contents)
        abs_json_path = tmp.name
        delete_files.append(abs_json_path)

    tmp_base_path = mkdtemp()
    delete_files.append(tmp_base_path)
    (Path(tmp_base_path) / ".nonadmin").touch()
    paths = install(abs_json_path, base_prefix=tmp_base_path)
    try:
        if action == "run_shortcut":
            if PLATFORM == "win":
                lnk = next(p for p in paths if p.suffix == ".lnk")
                assert lnk.is_file()
                os.startfile(lnk)
                output = _poll_for_file_contents(output_file)
            else:
                if PLATFORM == "linux":
                    desktop = next(p for p in paths if p.suffix == ".desktop")
                    with open(desktop) as f:
                        for line in f:
                            if line.startswith("Exec="):
                                cmd = shlex.split(line.split("=", 1)[1].strip())
                                break
                        else:
                            raise ValueError("Didn't find Exec line")
                elif PLATFORM == "osx":
                    app_location = paths[0]
                    executable = next(
                        p
                        for p in (app_location / "Contents" / "MacOS").iterdir()
                        if not p.name.endswith("-script")
                    )
                    cmd = [str(executable)]
                process = logged_run(cmd, check=True)
                output = process.stdout
        elif action is not None:
            if action == "open_file":
                assert file_to_open is not None
                with NamedTemporaryFile(suffix=file_to_open, delete=False) as f:
                    # file cannot be empty; otherwise mimetype detection fails on Linux
                    f.write(b"1234")
                delete_files.append(f.name)
                arg = f.name
            elif action == "open_url":
                assert url_to_open is not None
                arg = url_to_open
            app_location = paths[0]
            cmd = {
                "linux": ["xdg-open"],
                "osx": ["open"],
                "win": ["cmd", "/C", "start"],
            }[PLATFORM]
            process = logged_run([*cmd, arg], check=True)
            output = _poll_for_file_contents(output_file)
    finally:
        if paths:
            delete_files += list(paths)
        if remove_after:
            remove(abs_json_path, base_prefix=tmp_base_path)
        if PLATFORM == "osx" and action in ("open_file", "open_url"):
            _lsregister(
                "-kill",
                "-r",
                "-domain",
                "local",
                "-domain",
                "user",
                "-domain",
                "system",
            )
            sleep(5)
            if "menuinst" in _lsregister("-dump", log=False).stdout:
                warnings.warn(
                    "menuinst still registered with LaunchServices! "
                    "This usually fixes itself after a couple minutes. "
                    "Run '/System/Library/Frameworks/CoreServices.framework/Frameworks/"
                    "LaunchServices.framework/Support/lsregister -dump | grep menuinst' "
                    "to double check."
                )

    if expected_output is not None:
        assert output.strip() == expected_output

    return abs_json_path, paths, tmp_base_path, output


def test_install_prefix(delete_files):
    check_output_from_shortcut(delete_files, "sys-prefix.json", expected_output=sys.prefix)


def test_install_remove(tmp_path, delete_files):
    metadata = DATA / "jsons" / "sys-prefix.json"
    (tmp_path / ".nonadmin").touch()
    paths = set(install(metadata, target_prefix=tmp_path, base_prefix=tmp_path, _mode="user"))
    delete_files.extend(paths)
    files_found = set(filter(lambda x: x.exists(), paths))
    assert files_found == paths
    if PLATFORM != "osx":
        metadata_2 = json.loads(metadata.read_text())
        metadata_2["menu_items"][0]["name"] = "Sys.Prefix.2"
        paths_2 = set(
            install(metadata_2, target_prefix=tmp_path, base_prefix=tmp_path, _mode="user")
        )
        delete_files.extend(paths_2)
        files_found = set(filter(lambda x: x.exists(), paths_2.union(paths)))
        assert files_found == paths_2.union(paths)
        remove(metadata_2, target_prefix=tmp_path, base_prefix=tmp_path, _mode="user")
        files_found = set(filter(lambda x: x.exists(), paths_2.union(paths)))
        assert files_found == paths
    remove(metadata, target_prefix=tmp_path, base_prefix=tmp_path, _mode="user")
    files_found = set(filter(lambda x: x.exists(), paths))
    assert files_found == set()


def test_overwrite_existing_shortcuts(delete_files, caplog):
    """Test that overwriting shortcuts works without errors by running installation twice."""
    check_output_from_shortcut(
        delete_files,
        "precommands.json",
        remove_after=False,
    )
    if PLATFORM == "osx":
        with pytest.raises(RuntimeError):
            check_output_from_shortcut(
                delete_files,
                "precommands.json",
                remove_after=True,
            )
    else:
        caplog.clear()
        check_output_from_shortcut(
            delete_files,
            "precommands.json",
            remove_after=True,
        )
        assert any(line.startswith("Overwriting existing") for line in caplog.messages)


@pytest.mark.skipif(PLATFORM == "osx", reason="No menu names on MacOS")
def test_placeholders_in_menu_name(delete_files):
    _, paths, tmp_base_path, _ = check_output_from_shortcut(
        delete_files,
        "sys-prefix.json",
        expected_output=sys.prefix,
        remove_after=False,
    )
    if PLATFORM == "win":
        for path in paths:
            if path.suffix == ".lnk" and "Start Menu" in path.parts:
                assert path.parent.name == f"Sys.Prefix {Path(tmp_base_path).name}"
                break
        else:
            raise AssertionError("Didn't find Start Menu")
    elif PLATFORM == "linux":
        config_directory = Path(os.environ.get("XDG_CONFIG_HOME", "~/.config")).expanduser()
        desktop_directory = (
            Path(os.environ.get("XDG_DATA_HOME", "~/.local/share")).expanduser()
            / "desktop-directories"
        )
        menu_config_location = config_directory / "menus" / "applications.menu"
        rendered_name = f"Sys.Prefix {Path(tmp_base_path).name}"
        slugified_name = slugify(rendered_name)

        entry_file = desktop_directory / f"{slugified_name}.directory"
        assert entry_file.exists()
        entry = entry_file.read_text().splitlines()
        assert f"Name={rendered_name}" in entry

        tree = ElementTree.parse(menu_config_location)
        root = tree.getroot()
        assert rendered_name in [elt.text for elt in root.findall("Menu/Name")]
        assert f"{slugified_name}.directory" in [
            elt.text for elt in root.findall("Menu/Directory")
        ]
        assert rendered_name in [elt.text for elt in root.findall("Menu/Include/Category")]


def test_precommands(delete_files):
    check_output_from_shortcut(
        delete_files, "precommands.json", expected_output="rhododendron and bees"
    )


@pytest.mark.skipif(PLATFORM != "osx", reason="macOS only")
def test_entitlements(delete_files):
    json_path, paths, *_ = check_output_from_shortcut(
        delete_files, "entitlements.json", remove_after=False, expected_output="entitlements"
    )
    # verify signature
    app_dir = next(p for p in paths if p.name.endswith(".app"))
    subprocess.check_call(["/usr/bin/codesign", "--verbose", "--verify", str(app_dir)])

    launcher = next(
        p for p in (app_dir / "Contents" / "MacOS").iterdir() if not p.name.endswith("-script")
    )
    subprocess.check_call(["/usr/bin/codesign", "--verbose", "--verify", str(launcher)])

    for path in app_dir.rglob("Info.plist"):
        plist = plistlib.loads(path.read_bytes())
        assert plist
        assert "entitlements" not in plist
        break
    else:
        raise AssertionError("Didn't find Info.plist")

    for path in app_dir.rglob("Entitlements.plist"):
        plist = plistlib.loads(path.read_bytes())
        assert plist
        break
    else:
        raise AssertionError("Didn't find Entitlements.plist")

    remove(json_path)


@pytest.mark.skipif(PLATFORM != "osx", reason="macOS only")
def test_no_entitlements_no_signature(delete_files):
    json_path, paths, *_ = check_output_from_shortcut(
        delete_files, "sys-prefix.json", remove_after=False, expected_output=sys.prefix
    )
    app_dir = next(p for p in paths if p.name.endswith(".app"))
    launcher = next(
        p for p in (app_dir / "Contents" / "MacOS").iterdir() if not p.name.endswith("-script")
    )
    with pytest.raises(subprocess.CalledProcessError):
        subprocess.check_call(["/usr/bin/codesign", "--verbose", "--verify", str(app_dir)])
    with pytest.raises(subprocess.CalledProcessError):
        subprocess.check_call(["/usr/bin/codesign", "--verbose", "--verify", str(launcher)])
    remove(json_path)


@pytest.mark.skipif(PLATFORM != "osx", reason="macOS only")
def test_info_plist(delete_files):
    json_path, paths, *_ = check_output_from_shortcut(
        delete_files, "entitlements.json", remove_after=False, expected_output="entitlements"
    )
    app_dir = next(p for p in paths if p.name.endswith(".app"))

    for path in app_dir.rglob("Info.plist"):
        plist = plistlib.loads(path.read_bytes())
        assert plist
        break
    else:
        raise AssertionError("Didn't find file")

    assert plist["LSEnvironment"]["example_var"] == "example_value"

    remove(json_path)


@pytest.mark.skipif(PLATFORM != "osx", reason="macOS only")
def test_osx_symlinks(delete_files):
    json_path, paths, _, output = check_output_from_shortcut(
        delete_files, "osx_symlinks.json", remove_after=False
    )
    app_dir = next(p for p in paths if p.name.endswith(".app"))
    symlinked_python = app_dir / "Contents" / "Resources" / "python"
    assert output.strip() == str(symlinked_python)
    assert symlinked_python.resolve() == (Path(DEFAULT_PREFIX) / "bin" / "python").resolve()
    remove(json_path)


def _dump_ls_services():
    lsservicesplist = Path(
        os.environ["HOME"],
        "Library/Preferences/com.apple.LaunchServices/com.apple.launchservices.secure.plist",
    )
    plist = plistlib.loads(lsservicesplist.read_bytes())
    print(json.dumps(plist, indent=2))


@pytest.mark.skipif("CI" not in os.environ, reason="Only run on CI. Export CI=1 to run locally.")
def test_file_type_association(delete_files):
    test_file = "test.menuinst"
    *_, output = check_output_from_shortcut(
        delete_files,
        "file_types.json",
        action="open_file",
        file_to_open=test_file,
    )
    assert output.strip().endswith(test_file)


@pytest.mark.skipif(sys.platform != "darwin", reason="Only run on macOS")
@pytest.mark.skipif("CI" not in os.environ, reason="Only run on CI. Export CI=1 to run locally.")
def test_file_type_association_no_event_handler(delete_files, request):
    test_file = "test.menuinst-no-event-handler"
    abs_json_path, paths, tmp_base_path, _ = check_output_from_shortcut(
        delete_files,
        "file_types_no_event_handler.json",
        action=None,
        file_to_open=test_file,
        remove_after=False,
    )
    request.addfinalizer(lambda: remove(abs_json_path, base_prefix=tmp_base_path))
    app_dir = next(p for p in paths if p.name.endswith(".app"))
    info = app_dir / "Contents" / "Info.plist"
    plist = plistlib.loads(info.read_bytes())
    cf_bundle_type_name = "org.conda.menuinst.filetype-example-no-event-handler"
    assert plist["CFBundleDocumentTypes"][0]["CFBundleTypeName"] == cf_bundle_type_name


@pytest.mark.skipif("CI" not in os.environ, reason="Only run on CI. Export CI=1 to run locally.")
def test_url_protocol_association(delete_files):
    url = "menuinst://test/"
    check_output_from_shortcut(
        delete_files,
        "url_protocols.json",
        action="open_url",
        url_to_open=url,
        expected_output=url,
    )


@pytest.mark.skipif(PLATFORM != "win", reason="Windows only")
def test_windows_terminal_profiles(tmp_path, run_as_user):
    settings_file = Path(
        tmp_path, "localappdata", "Microsoft", "Windows Terminal", "settings.json"
    )
    settings_file.parent.mkdir(parents=True)
    (tmp_path / ".nonadmin").touch()
    metadata_file = DATA / "jsons" / "windows-terminal.json"
    install(metadata_file, target_prefix=tmp_path, base_prefix=tmp_path)
    try:
        settings = json.loads(settings_file.read_text())
        profiles = {
            profile.get("name", ""): profile.get("commandline", "")
            for profile in settings.get("profiles", {}).get("list", [])
        }
        assert profiles.get("A Terminal") == "testcommand_a.exe"
        assert "B" not in profiles
    except Exception as exc:
        remove(metadata_file, target_prefix=tmp_path, base_prefix=tmp_path)
        raise exc
    else:
        remove(metadata_file, target_prefix=tmp_path, base_prefix=tmp_path)


@pytest.mark.parametrize("target_env_is_base", (True, False))
def test_name_dictionary(target_env_is_base):
    tmp_base_path = mkdtemp()
    tmp_target_path = tmp_base_path if target_env_is_base else mkdtemp()
    (Path(tmp_base_path) / ".nonadmin").touch()
    if not target_env_is_base:
        (Path(tmp_target_path) / ".nonadmin").touch()
    abs_json_path = DATA / "jsons" / "menu-name.json"
    menu_items = install(abs_json_path, target_prefix=tmp_target_path, base_prefix=tmp_base_path)
    try:
        if PLATFORM == "linux":
            expected = {
                "package_a" if target_env_is_base else "package_a-not-in-base",
                "package_b",
                "package",
            }
        else:
            expected = {
                "A" if target_env_is_base else "A_not_in_base",
                "B",
            }
            if PLATFORM == "win":
                expected.update(["Package"])
        item_names = {item.stem for item in menu_items}
        assert item_names == expected
    finally:
        remove(abs_json_path, target_prefix=tmp_target_path, base_prefix=tmp_base_path)


def test_vars_in_working_dir(tmp_path, monkeypatch, delete_files):
    if PLATFORM == "win":
        expected_directory = Path(os.environ["TEMP"], "working_dir_test")
    elif PLATFORM == "osx":
        expected_directory = Path(os.environ["TMPDIR"], "working_dir_test")
    else:
        # Linux often does not have an environment variable for the tmp directory
        monkeypatch.setenv("TMP", "/tmp")
        expected_directory = Path("/tmp/working_dir_test")
    delete_files.append(expected_directory)
    datafile = str(DATA / "jsons" / "working-dir.json")
    try:
        install(datafile, base_prefix=tmp_path, target_prefix=tmp_path)
        assert expected_directory.exists()
    finally:
        remove(datafile, base_prefix=tmp_path, target_prefix=tmp_path)



import os

import pytest


@pytest.mark.skipif(os.name != "nt", reason="Windows only")
def test_import_paths():
    """Imports used by conda <=23.7.2. Ensure they still work."""
    import menuinst._legacy.cwp  # noqa: F401
    import menuinst._legacy.main  # noqa: F401
    import menuinst._legacy.utils  # noqa: F401
    import menuinst._legacy.win32  # noqa: F401
    from menuinst import install  # noqa: F401
    from menuinst.knownfolders import FOLDERID, get_folder_path  # noqa: F401
    from menuinst.win32 import dirs_src  # noqa: F401
    from menuinst.win_elevate import isUserAdmin, runAsAdmin  # noqa: F401
    from menuinst.winshortcut import create_shortcut  # noqa: F401



import os
import time
from pathlib import Path

import pytest

registry = pytest.importorskip("menuinst.platforms.win_utils.registry")


def test_file_extensions(tmp_path: Path, request):
    """
    We will register a custom, random extension.
    The command will only echo the input path to a known output location.
    After "opening" an empty file with that extension, the output file
    should contain the file path we opened!

    We then clean up and make sure that opening that file doesn't
    create the output file.
    """
    name = str(hash(str(tmp_path)))[:6]
    extension = f".menuinst-{name}"
    identifier = f"menuinst.assoc.menuinst-{name}"

    def cleanup():
        # This key is not normally cleaned up because another programs might
        # be using it, but since we know these are synthetic and made up,
        # we try not to pollute the registry too much
        registry._reg_exe("delete", fr"HKCU\Software\Classes\.menuinst-{name}")

    request.addfinalizer(cleanup)

    registry.register_file_extension(
        extension=extension,
        identifier=identifier,
        command=fr'cmd.exe /Q /D /V:ON /C "echo %1>{tmp_path}\output.txt"',
        mode="user",
    )
    input_file = tmp_path / f"input.menuinst-{name}"
    output_file = tmp_path / "output.txt"
    try:
        input_file.touch()
        os.startfile(input_file)
        t0 = time.time()
        while time.time() - t0 <= 10:  # wait 10 seconds
            try:
                assert output_file.read_text().strip() == str(input_file)
            except IOError:
                time.sleep(1)
        if not output_file.exists():
            raise AssertionError("Output file was never created")
    finally:
        registry.unregister_file_extension(extension=extension, identifier=identifier, mode="user")

    output_file.unlink()
    os.startfile(input_file)  # this will raise UI if not headless, ignore
    t0 = time.time()
    while time.time() - t0 <= 3:  # wait up to 3 seconds
        time.sleep(1)
        assert not output_file.exists()


def test_protocols(tmp_path):
    """
    We will register a custom, random protocol.
    The command will only echo the input path to a known output location.
    After "opening" a fake URL with that protocol, the output file
    should contain the file path we opened!

    We then clean up and make sure that opening that file doesn't
    create the output file.
    """
    name = str(hash(str(tmp_path)))[:6]
    registry.register_url_protocol(
        protocol=f"menuinst-{name}",
        command=fr'cmd.exe /Q /D /V:ON /C "echo %1>{tmp_path}\output.txt"',
        identifier=f"menuinst.protocol.menuinst-{name}",
        mode="user",
    )
    input_url = f"menuinst-{name}://fake-value"
    output_file = tmp_path / "output.txt"
    os.startfile(input_url)
    try:
        t0 = time.time()
        while time.time() - t0 <= 10:  # wait 10 seconds
            try:
                assert output_file.read_text().strip().rstrip("/") == input_url.rstrip("/")
            except IOError:
                time.sleep(1)
        if not output_file.exists():
            raise AssertionError("Output file was never created")
    finally:
        registry.unregister_url_protocol(
            protocol=f"menuinst-{name}",
            identifier=f"menuinst.protocol.menuinst-{name}",
            mode="user",
        )

    output_file.unlink()
    os.startfile(input_url)  # this will raise UI if not headless, ignore
    t0 = time.time()
    while time.time() - t0 <= 3:  # wait up to 3 seconds
        time.sleep(1)
        assert not output_file.exists()



"""
"""

import json
import os
import sys
import warnings
from logging import getLogger
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union

from .platforms import Menu, MenuItem
from .utils import DEFAULT_BASE_PREFIX, DEFAULT_PREFIX, _UserOrSystem, elevate_as_needed

log = getLogger(__name__)


__all__ = [
    "install",
    "remove",
    "install_all",
    "remove_all",
]


def _load(
    metadata_or_path: Union[os.PathLike, dict],
    target_prefix: Optional[os.PathLike] = None,
    base_prefix: Optional[os.PathLike] = None,
    _mode: _UserOrSystem = "user",
) -> Tuple[Menu, List[MenuItem]]:
    target_prefix = target_prefix or DEFAULT_PREFIX
    base_prefix = base_prefix or DEFAULT_BASE_PREFIX
    if isinstance(metadata_or_path, (str, Path)):
        with open(metadata_or_path) as f:
            metadata = json.load(f)
    else:
        metadata = metadata_or_path
    menu = Menu(metadata["menu_name"], target_prefix, base_prefix, _mode)
    menu_items = [MenuItem(menu, item) for item in metadata["menu_items"]]
    return menu, menu_items


@elevate_as_needed
def install(
    metadata_or_path: Union[os.PathLike, dict],
    *,
    target_prefix: Optional[os.PathLike] = None,
    base_prefix: Optional[os.PathLike] = None,
    _mode: _UserOrSystem = "user",
) -> List[os.PathLike]:
    target_prefix = target_prefix or DEFAULT_PREFIX
    base_prefix = base_prefix or DEFAULT_BASE_PREFIX
    menu, menu_items = _load(metadata_or_path, target_prefix, base_prefix, _mode)
    if not any(item.enabled_for_platform() for item in menu_items):
        warnings.warn(f"Metadata for {menu.name} is not enabled for {sys.platform}")
        return ()

    paths = []
    paths += menu.create()
    for menu_item in menu_items:
        paths += menu_item.create()

    return paths


@elevate_as_needed
def remove(
    metadata_or_path: Union[os.PathLike, dict],
    *,
    target_prefix: Optional[os.PathLike] = None,
    base_prefix: Optional[os.PathLike] = None,
    _mode: _UserOrSystem = "user",
) -> List[os.PathLike]:
    target_prefix = target_prefix or DEFAULT_PREFIX
    base_prefix = base_prefix or DEFAULT_BASE_PREFIX
    menu, menu_items = _load(metadata_or_path, target_prefix, base_prefix, _mode)
    if not any(item.enabled_for_platform() for item in menu_items):
        warnings.warn(f"Metadata for {menu.name} is not enabled for {sys.platform}")
        return ()

    paths = []
    for menu_item in menu_items:
        paths += menu_item.remove()
    paths += menu.remove()

    return paths


@elevate_as_needed
def install_all(
    *,
    target_prefix: Optional[os.PathLike] = None,
    base_prefix: Optional[os.PathLike] = None,
    filter: Union[Callable, None] = None,
    _mode: _UserOrSystem = "user",
) -> List[List[os.PathLike]]:
    target_prefix = target_prefix or DEFAULT_PREFIX
    base_prefix = base_prefix or DEFAULT_BASE_PREFIX
    return _process_all(install, target_prefix, base_prefix, filter, _mode)


@elevate_as_needed
def remove_all(
    *,
    target_prefix: Optional[os.PathLike] = None,
    base_prefix: Optional[os.PathLike] = None,
    filter: Union[Callable, None] = None,
    _mode: _UserOrSystem = "user",
) -> List[List[Union[str, os.PathLike]]]:
    target_prefix = target_prefix or DEFAULT_PREFIX
    base_prefix = base_prefix or DEFAULT_BASE_PREFIX
    return _process_all(remove, target_prefix, base_prefix, filter, _mode)


def _process_all(
    function: Callable,
    target_prefix: Optional[os.PathLike] = None,
    base_prefix: Optional[os.PathLike] = None,
    filter: Union[Callable, None] = None,
    _mode: _UserOrSystem = "user",
) -> List[Any]:
    target_prefix = target_prefix or DEFAULT_PREFIX
    base_prefix = base_prefix or DEFAULT_BASE_PREFIX
    jsons = (Path(target_prefix) / "Menu").glob("*.json")
    results = []
    for path in jsons:
        if filter is not None and filter(path):
            results.append(function(path, target_prefix, base_prefix, _mode))
    return results


_api_remove = remove  # alias to prevent shadowing in the function below


def _install_adapter(
    path: os.PathLike, remove: bool = False, prefix: os.PathLike = DEFAULT_PREFIX, **kwargs
):
    """
    This function is only here as a legacy adapter for menuinst v1.x.
    Please use `menuinst.api` functions instead.
    """
    if os.name == "nt":
        path = path.replace("/", "\\")
    json_path = os.path.join(prefix, path)
    with open(json_path) as f:
        metadata = json.load(f)
    if "$id" not in metadata:  # old style JSON
        from ._legacy import install as _legacy_install

        if os.name == "nt":
            kwargs.setdefault("root_prefix", kwargs.pop("base_prefix", DEFAULT_BASE_PREFIX))
            if kwargs["root_prefix"] is None:
                kwargs["root_prefix"] = DEFAULT_BASE_PREFIX
            _legacy_install(json_path, remove=remove, prefix=prefix, **kwargs)
        else:
            log.warn(
                "menuinst._legacy is only supported on Windows. "
                "Switch to the new-style menu definitions "
                "for cross-platform compatibility."
            )
    else:
        # patch kwargs to reroute root_prefix to base_prefix
        kwargs.setdefault("base_prefix", kwargs.pop("root_prefix", DEFAULT_BASE_PREFIX))
        if kwargs["base_prefix"] is None:
            kwargs["base_prefix"] = DEFAULT_BASE_PREFIX
        if remove:
            _api_remove(metadata, target_prefix=prefix, **kwargs)
        else:
            install(metadata, target_prefix=prefix, **kwargs)



import os

from menuinst.utils import _test_elevation, elevate_as_needed, user_is_admin


def test_elevation(tmp_path, capfd):
    os.environ["MENUINST_TEST"] = "TEST"
    if os.name == "nt":
        on_ci = os.environ.get("CI")
        is_admin = user_is_admin()
        if not on_ci:
            # Windows runners on GHA always run as admin
            assert not is_admin

        _test_elevation(target_prefix=str(tmp_path), base_prefix=str(tmp_path))
        output = (tmp_path / "_test_output.txt").read_text().strip()
        if on_ci:
            assert output.endswith("env_var: TEST _mode: user")
        else:
            assert output.endswith("user_is_admin(): False env_var: TEST _mode: user")

        elevate_as_needed(_test_elevation)(target_prefix=str(tmp_path), base_prefix=str(tmp_path))
        output = (tmp_path / "_test_output.txt").read_text().strip()
        if on_ci:
            assert output.endswith("env_var: TEST _mode: system")
        else:
            assert output.endswith("user_is_admin(): True env_var: TEST _mode: system")
    else:
        assert not user_is_admin()  # We need to start as a non-root user

        _test_elevation(str(tmp_path))
        assert capfd.readouterr().out.strip() == "user_is_admin(): False env_var: TEST _mode: user"

        # make tmp_path not writable by the current user to force elevation
        tmp_path.chmod(0o500)
        elevate_as_needed(_test_elevation)(target_prefix=str(tmp_path), base_prefix=str(tmp_path))
        assert (
            capfd.readouterr().out.strip() == "user_is_admin(): True env_var: TEST _mode: system"
        )
        assert not (tmp_path / ".nonadmin").exists()

        # restore permissions
        tmp_path.chmod(0o700)
        elevate_as_needed(_test_elevation)(target_prefix=str(tmp_path), base_prefix=str(tmp_path))
        assert capfd.readouterr().out.strip() == "user_is_admin(): False env_var: TEST _mode: user"
        assert (tmp_path / ".nonadmin").exists()



r"""
Utilities for Windows Registry manipulation

Notes:

winreg.SetValue -> sets _keys_ with default ("") values
winreg.SetValueEx -> sets values with named contents

This is important when the argument has backslashes.
SetValue will process the backslashes into a path of keys
SetValueEx will create a value with name "path\with\keys"


Mnemonic: SetValueEx for "excalars" (scalars, named values)
"""

import winreg
from logging import getLogger

from ...utils import logged_run

log = getLogger(__name__)


def _reg_exe(*args, check=True):
    return logged_run(["reg.exe", *args, "/f"], check=True)


def register_file_extension(extension, identifier, command, icon=None, mode="user"):
    """
    We want to achieve this. Entries ending in / denote keys; no trailing / means named value.
    If the item has a value attached to it, it's written after the : symbol.

    HKEY_*/ # current user or local machine
      Software/
        Classes/
          .<extension>/
            OpenWithProgids/
              <extension-handler>
          ...
          <extension-handler>/: "a description of the file being handled"
            DefaultIcon: "path to the app icon"
            shell/
              open/
                command/: "the command to be executed when opening a file with this extension"
    """
    with winreg.OpenKeyEx(
        (
            winreg.HKEY_LOCAL_MACHINE if mode == "system" else winreg.HKEY_CURRENT_USER  # HKLM
        ),  # HKCU
        r"Software\Classes",
    ) as key:
        # First we associate an extension with a handler
        winreg.SetValueEx(
            winreg.CreateKey(key, fr"{extension}\OpenWithProgids"),
            identifier,
            0,
            winreg.REG_SZ,
            "",  # presence of the key is enough
        )
        log.debug("Created registry entry for extension '%s'", extension)

        # Now we register the handler
        handler_desc = f"{extension} {identifier} handler"
        winreg.SetValue(key, identifier, winreg.REG_SZ, handler_desc)
        log.debug("Created registry entry for handler '%s'", identifier)

        # and set the 'open' command
        subkey = rf"{identifier}\shell\open\command"
        # Use SetValue to create subkeys as necessary
        winreg.SetValue(key, subkey, winreg.REG_SZ, command)
        log.debug("Created registry entry for command '%s'", command)

        if icon:
            subkey = winreg.OpenKey(key, identifier, access=winreg.KEY_SET_VALUE)
            winreg.SetValueEx(subkey, "DefaultIcon", 0, winreg.REG_SZ, icon)
            log.debug("Created registry entry for icon '%s'", icon)

        # TODO: We can add contextual menu items too
        # via f"{handler_key}\shell\<Command Title>\command"


def unregister_file_extension(extension, identifier, mode="user"):
    root, root_str = (
        (winreg.HKEY_LOCAL_MACHINE, "HKLM")
        if mode == "system"
        else (winreg.HKEY_CURRENT_USER, "HKCU")
    )
    _reg_exe("delete", fr"{root_str}\Software\Classes\{identifier}")

    try:
        with winreg.OpenKey(
            root, fr"Software\Classes\{extension}\OpenWithProgids", 0, winreg.KEY_ALL_ACCESS
        ) as key:
            try:
                winreg.QueryValueEx(key, identifier)
            except FileNotFoundError:
                log.debug(
                    "Handler '%s' is not associated with extension '%s'", identifier, extension
                )
            else:
                winreg.DeleteValue(key, identifier)
    except Exception as exc:
        log.exception("Could not check key '%s' for deletion", extension, exc_info=exc)
        raise


def register_url_protocol(protocol, command, identifier=None, icon=None, mode="user"):
    if mode == "system":
        key = winreg.CreateKey(winreg.HKEY_CLASSES_ROOT, protocol)
    else:
        key = winreg.CreateKey(winreg.HKEY_CURRENT_USER, fr"Software\Classes\{protocol}")
    with key:
        winreg.SetValueEx(key, "", 0, winreg.REG_SZ, f"URL:{protocol.title()}")
        winreg.SetValueEx(key, "URL Protocol", 0, winreg.REG_SZ, "")
        # SetValue creates sub keys when slashes are present;
        # SetValueEx creates a value with backslashes - we don't want that here
        winreg.SetValue(key, r"shell\open\command", winreg.REG_SZ, command)
        if icon:
            winreg.SetValueEx(key, "DefaultIcon", 0, winreg.REG_SZ, icon)
        if identifier:
            # We add this one value for traceability; not required
            winreg.SetValueEx(key, "_menuinst", 0, winreg.REG_SZ, identifier)


def unregister_url_protocol(protocol, identifier=None, mode="user"):
    if mode == "system":
        key_tuple = winreg.HKEY_CLASSES_ROOT, protocol
        key_str = fr"HKCR\{protocol}"
    else:
        key_tuple = winreg.HKEY_CURRENT_USER, fr"Software\Classes\{protocol}"
        key_str = fr"HKCU\Software\Classes\{protocol}"
    try:
        with winreg.OpenKey(*key_tuple) as key:
            value, _ = winreg.QueryValueEx(key, "_menuinst")
            delete = identifier is None or value == identifier
    except OSError as exc:
        log.exception("Could not check key %s for deletion", protocol, exc_info=exc)
        return

    if delete:
        _reg_exe("delete", key_str, check=False)



# Copyright (c) 2008-2011 by Enthought, Inc.
# Copyright (c) 2013-2017 Continuum Analytics, Inc.
# All rights reserved.

from __future__ import absolute_import, unicode_literals

import ctypes
import locale
import logging
import os
import sys
from os.path import exists, isdir, join

from ..platforms.win_utils.knownfolders import dirs_src, folder_path
from ..platforms.win_utils.winshortcut import create_shortcut
from ..utils import DEFAULT_BASE_PREFIX
from .utils import rm_empty_dir, rm_rf

# This allows debugging installer issues using DebugView from Microsoft.
OutputDebugString = ctypes.windll.kernel32.OutputDebugStringW
OutputDebugString.argtypes = [ctypes.c_wchar_p]


class DbgViewHandler(logging.Handler):
    def emit(self, record):
        OutputDebugString(self.format(record))


logger = logging.getLogger("menuinst_win32")
logger.setLevel(logging.DEBUG)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.WARNING)
dbgview = DbgViewHandler()
dbgview.setLevel(logging.DEBUG)
logger.addHandler(dbgview)
logger.addHandler(stream_handler)


def quoted(s):
    """
    quotes a string if necessary.
    """
    # strip any existing quotes
    s = s.strip(u'"')
    # don't add quotes for minus or leading space
    if s[0] in (u'-', u' '):
        return s
    if u' ' in s or u'/' in s:
        return u'"%s"' % s
    else:
        return s


def ensure_pad(name, pad="_"):
    """

    Examples:
        >>> ensure_pad('conda')
        '_conda_'

    """
    if not name or name[0] == name[-1] == pad:
        return name
    else:
        return "%s%s%s" % (pad, name, pad)


def to_unicode(var, codec=locale.getpreferredencoding()):
    if not codec:
        codec = "utf-8"
    if hasattr(var, "decode"):
        var = var.decode(codec)
    return var


def to_bytes(var, codec=locale.getpreferredencoding()):
    if isinstance(var, bytes):
        return var
    if not codec:
        codec = "utf-8"
    if hasattr(var, "encode"):
        var = var.encode(codec)
    return var


unicode_root_prefix = to_unicode(DEFAULT_BASE_PREFIX)
if u'\\envs\\' in unicode_root_prefix:
    logger.warning('menuinst called from non-root env %s', unicode_root_prefix)


def substitute_env_variables(text, dir):
    # When conda is using Menuinst, only the root conda installation ever
    # calls menuinst.  Thus, these calls to sys refer to the root conda
    # installation, NOT the child environment
    py_major_ver = sys.version_info[0]
    py_bitness = 8 * tuple.__itemsize__

    env_prefix = to_unicode(dir['prefix'])
    root_prefix = to_unicode(dir['root_prefix'])
    text = to_unicode(text)
    env_name = to_unicode(dir['env_name'])

    for a, b in (
        (u'${PREFIX}', env_prefix),
        (u'${ROOT_PREFIX}', root_prefix),
        (u'${DISTRIBUTION_NAME}', os.path.split(root_prefix)[-1]),
        (
            u'${PYTHON_SCRIPTS}',
            os.path.normpath(join(env_prefix, u'Scripts')).replace(u"\\", u"/"),
        ),
        (u'${MENU_DIR}', join(env_prefix, u'Menu')),
        (u'${PERSONALDIR}', dir['documents']),
        (u'${USERPROFILE}', dir['profile']),
        (u'${ENV_NAME}', env_name),
        (u'${PY_VER}', u'%d' % (py_major_ver)),
        (u'${PLATFORM}', u"(%s-bit)" % py_bitness),
    ):
        if b:
            text = text.replace(a, b)
    return text


class Menu(object):
    def __init__(
        self,
        name,
        prefix=unicode_root_prefix,
        env_name=u"",
        mode=None,
        root_prefix=unicode_root_prefix,
    ):
        """
        Prefix is the system prefix to be used -- this is needed since
        there is the possibility of a different Python's packages being managed.
        """

        # bytestrings passed in need to become unicode
        self.prefix = to_unicode(prefix)
        self.root_prefix = to_unicode(root_prefix)
        used_mode = (
            mode if mode else ('user' if exists(join(self.prefix, u'.nonadmin')) else 'system')
        )
        logger.debug(
            "Menu: name: '%s', prefix: '%s', env_name: '%s', mode: '%s', used_mode: '%s', root_prefix: '%s'"  # noqa
            % (name, self.prefix, env_name, mode, used_mode, root_prefix)
        )
        try:
            self.set_dir(name, self.prefix, env_name, used_mode, root_prefix)
        except WindowsError:
            # We get here if we aren't elevated.  This is different from
            #   permissions: a user can have permission, but elevation is still
            #   required.  If the process isn't elevated, we get the
            #   WindowsError
            if 'user' in dirs_src and used_mode == 'system':
                logger.warn(
                    "Insufficient permissions to write menu folder.  "
                    "Falling back to user location"
                )
                try:
                    self.set_dir(name, self.prefix, env_name, 'user')
                except:  # noqa
                    pass
            else:
                logger.fatal("Unable to create AllUsers menu folder")

    def set_dir(self, name, prefix, env_name, mode, root_prefix):
        self.mode = mode
        self.dir = dict()
        # I have chickened out on allowing check_other_mode. Really there needs
        # to be 3 distinct cases that 'menuinst' cares about:
        # priv-user doing system install
        # priv-user doing user-only install
        # non-priv-user doing user-only install
        # (priv-user only exists in an AllUsers installation).
        check_other_mode = False
        for k, v in dirs_src[mode].items():
            # We may want to cache self.dir to some files, one for AllUsers
            # (system) installs and one for each subsequent user install?
            self.dir[k] = folder_path(mode, check_other_mode, k)
        self.dir['prefix'] = prefix
        self.dir['root_prefix'] = root_prefix
        self.dir['env_name'] = env_name
        folder_name = substitute_env_variables(name, self.dir)
        self.path = join(self.dir["start"], folder_name)
        self.create()

    def create(self):
        if not isdir(self.path):
            os.mkdir(self.path)

    def remove(self):
        rm_empty_dir(self.path)


def extend_script_args(args, shortcut):
    try:
        args.append(shortcut['scriptargument'])
    except KeyError:
        pass
    try:
        args.extend(shortcut['scriptarguments'])
    except KeyError:
        pass


def quote_args(args):
    # cmd.exe /K or /C expects a single string argument and requires
    # doubled-up quotes when any sub-arguments have spaces:
    # https://stackoverflow.com/a/6378038/3257826
    if (
        len(args) > 2
        and ("CMD.EXE" in args[0].upper() or "%COMSPEC%" in args[0].upper())
        and (args[1].upper() == '/K' or args[1].upper() == '/C')
        and any(' ' in arg for arg in args[2:])
    ):
        args = [
            ensure_pad(args[0], '"'),  # cmd.exe
            args[1],  # /K or /C
            '"%s"' % (' '.join(ensure_pad(arg, '"') for arg in args[2:])),  # double-quoted
        ]
    else:
        args = [quoted(arg) for arg in args]
    return args


class ShortCut(object):
    def __init__(self, menu, shortcut):
        self.menu = menu
        self.shortcut = shortcut

    def remove(self):
        self.create(remove=True)

    def create(self, remove=False):
        # Substitute env variables early because we may need to escape spaces in the value.
        args = []
        fix_win_slashes = [0]
        prefix = self.menu.prefix.replace('/', '\\')
        unicode_root_prefix = self.menu.root_prefix.replace('/', '\\')
        root_py = join(unicode_root_prefix, u"python.exe")
        root_pyw = join(unicode_root_prefix, u"pythonw.exe")
        env_py = join(prefix, u"python.exe")
        env_pyw = join(prefix, u"pythonw.exe")
        cwp_py = [root_py, join(unicode_root_prefix, u'cwp.py'), prefix, env_py]
        cwp_pyw = [root_pyw, join(unicode_root_prefix, u'cwp.py'), prefix, env_pyw]
        if "pywscript" in self.shortcut:
            args = cwp_pyw
            fix_win_slashes = [len(args)]
            args += self.shortcut["pywscript"].split()
        elif "pyscript" in self.shortcut:
            args = cwp_py
            fix_win_slashes = [len(args)]
            args += self.shortcut["pyscript"].split()
        elif "webbrowser" in self.shortcut:
            args = [root_pyw, '-m', 'webbrowser', '-t', self.shortcut['webbrowser']]
        elif "script" in self.shortcut:
            # It is unclear whether running through cwp.py is what we want here. In
            # the long term I would rather this was made an explicit choice.
            args = [root_py, join(unicode_root_prefix, u'cwp.py'), prefix]
            fix_win_slashes = [len(args)]
            args += self.shortcut["script"].split()
            extend_script_args(args, self.shortcut)
        elif "system" in self.shortcut:
            args = self.shortcut["system"].split()
            extend_script_args(args, self.shortcut)
        else:
            raise Exception("Nothing to do: %r" % self.shortcut)
        args = [substitute_env_variables(arg, self.menu.dir) for arg in args]
        for fws in fix_win_slashes:
            args[fws] = args[fws].replace('/', '\\')

        args = quote_args(args)

        cmd = args[0]
        args = args[1:]
        logger.debug('Shortcut cmd is %s, args are %s' % (cmd, args))
        workdir = self.shortcut.get('workdir', '')
        icon = self.shortcut.get('icon', '')

        workdir = substitute_env_variables(workdir, self.menu.dir)
        icon = substitute_env_variables(icon, self.menu.dir)

        # Fix up the '/' to '\'
        workdir = workdir.replace('/', '\\')
        icon = icon.replace('/', '\\')

        # Create the working directory if it doesn't exist
        if workdir:
            if not isdir(workdir):
                os.makedirs(workdir)
        else:
            workdir = '%HOMEPATH%'

        # Menu link
        dst_dirs = [self.menu.path]

        # Desktop link
        if self.shortcut.get('desktop'):
            dst_dirs.append(self.menu.dir['desktop'])

        # Quicklaunch link
        if self.shortcut.get('quicklaunch') and 'quicklaunch' in self.menu.dir:
            dst_dirs.append(self.menu.dir['quicklaunch'])

        name_suffix = (
            " ({})".format(self.menu.dir['env_name']) if self.menu.dir['env_name'] else ""
        )
        for dst_dir in dst_dirs:
            name = substitute_env_variables(self.shortcut['name'], self.menu.dir)
            dst = join(dst_dir, name + name_suffix + '.lnk')
            if remove:
                rm_rf(dst)
            else:
                # The API for the call to 'create_shortcut' has 3
                # required arguments (path, description and filename)
                # and 4 optional ones (args, working_dir, icon_path and
                # icon_index).
                create_shortcut(
                    u'' + cmd,
                    u'' + name + name_suffix,
                    u'' + dst,
                    u' '.join(arg for arg in args),
                    u'' + workdir,
                    u'' + icon,
                )



import os
import re
import shlex
import subprocess
import sys
import xml.etree.ElementTree as XMLTree
from contextlib import suppress
from functools import lru_cache, wraps
from logging import getLogger
from pathlib import Path
from typing import Callable, Iterable, Literal, Mapping, Optional, Sequence, Union
from unicodedata import normalize

logger = getLogger(__name__)
_TargetOrBase = Union[Literal["target"], Literal["base"]]
_UserOrSystem = Union[Literal["user"], Literal["system"]]


def _default_prefix(which: _TargetOrBase = "target"):
    """
    The prefixes in menuinst need to be handled with care.

    Conda installations that require superuser permissions need elevation for
    the creation of shortcuts. Constructor will leave a sentinel file to signal
    this. If a file `.nonadmin` is present in the 'base' environment (or root of
    the installation directory if conda is not present), superuser access is not
    needed.

    In order to check for this file, menuinst needs to track 'base_prefix'. For
    a regular 'conda' process, this should be `conda.base.context.context.root_prefix'.
    However, constructor also relies on a pyinstaller-frozen conda installation,
    'conda-standalone'. In these cases, 'sys.prefix' is set to temporary location
    of the extracted contents of the executable -- that's NOT the base installation!

    For those reasons, we handle the default prefix with 'sys.prefix' (or 'sys.base_prefix')
    as a last resort. The logic is:

    - If 'MENUINST_PREFIX' (or 'MENUINST_BASE_PREFIX') is an env var with a set value, use that.
    - If are already using conda, we get the context object and use those values.
    - If CONDA_PREFIX (or 'CONDA_ROOT_PREFIX') are available, use those
    - Last resort: use sys.prefix and sys.base_prefix

    This helps us pass a lot of CLI arguments back and forth.
    """
    base = which == "base"
    context = None
    if "conda" in sys.modules:
        with suppress(ImportError):
            from conda.base.context import context

    if base:
        prefix = os.environ.get("MENUINST_BASE_PREFIX")
        if prefix:
            return prefix
        if context:
            return context.root_prefix
        return os.environ.get("CONDA_ROOT_PREFIX", sys.base_prefix)
    # else
    prefix = os.environ.get("MENUINST_PREFIX")
    if prefix:
        return prefix
    if context:
        return context.target_prefix
    return os.environ.get("CONDA_PREFIX", sys.prefix)


DEFAULT_PREFIX = _default_prefix("target")
DEFAULT_BASE_PREFIX = _default_prefix("base")


def slugify(text: str):
    # Adapted from from django.utils.text.slugify
    # Copyright (c) Django Software Foundation and individual contributors.
    # All rights reserved.
    # Redistribution and use in source and binary forms, with or without modification,
    # are permitted provided that the following conditions are met:
    #     1. Redistributions of source code must retain the above copyright notice,
    #     this list of conditions and the following disclaimer.
    #     2. Redistributions in binary form must reproduce the above copyright
    #     notice, this list of conditions and the following disclaimer in the
    #     documentation and/or other materials provided with the distribution.
    #     3. Neither the name of Django nor the names of its contributors may be used
    #     to endorse or promote products derived from this software without
    #     specific prior written permission.
    # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    # ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
    # ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    # LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    # ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """
    Convert to ASCII, and spaces and underscores to hyphens.
    Remove characters that aren't alphanumerics, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace.
    """
    text = normalize("NFKD", text).encode("ascii", "ignore").decode("ascii")
    text = re.sub(r"[^\w\s-]", "", text).strip().lower()
    return re.sub(r"[_\-\s]+", "-", text)


def indent_xml_tree(elem, level=0):
    """
    adds whitespace to the tree, so that it results in a pretty printed tree
    """
    indentation = "    "  # 4 spaces, just like in Python!
    base_indentation = "\n" + level * indentation
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = base_indentation + indentation
        for e in elem:
            indent_xml_tree(e, level + 1)
            if not e.tail or not e.tail.strip():
                e.tail = base_indentation + indentation
        if not e.tail or not e.tail.strip():
            e.tail = base_indentation
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = base_indentation


def add_xml_child(parent: XMLTree.Element, tag: str, text: Optional[str] = None):
    """
    Add a child element of specified tag type to parent.
    The new child element is returned.
    """
    elem = XMLTree.SubElement(parent, tag)
    if text is not None:
        elem.text = text
    return elem


class WinLex:
    @classmethod
    def quote_args(cls, args: Sequence[str]):
        # cmd.exe /K or /C expects a single string argument and requires
        # doubled-up quotes when any sub-arguments have spaces:
        # https://stackoverflow.com/a/6378038/3257826
        if (
            len(args) > 2
            and ("CMD.EXE" in args[0].upper() or "%COMSPEC%" in args[0].upper())
            and (args[1].upper() == "/K" or args[1].upper() == "/C")
            and any(" " in arg for arg in args[2:])
        ):
            args = [
                cls.ensure_pad(args[0], '"'),  # cmd.exe
                args[1],  # /K or /C
                '"%s"' % (" ".join(cls.ensure_pad(arg, '"') for arg in args[2:])),  # double-quoted
            ]
        else:
            args = [cls.quote_string(arg) for arg in args]
        return args

    @classmethod
    def quote_string(cls, s: Sequence[str]):
        """
        quotes a string if necessary.
        """
        # strip any existing quotes
        s = s.strip('"')
        # don't add quotes for minus or leading space
        if s[0] in ("-", " "):
            return s
        if " " in s or "/" in s:
            return '"%s"' % s
        return s

    @classmethod
    def ensure_pad(cls, name: str, pad: str = "_"):
        """

        Examples:
            >>> ensure_pad('conda')
            '_conda_'

        """
        if not name or name[0] == name[-1] == pad:
            return name
        else:
            return "%s%s%s" % (pad, name, pad)


class UnixLex:
    @classmethod
    def quote_args(cls, args: Sequence[str]) -> Sequence[str]:
        return [cls.quote_string(a) for a in args]

    @classmethod
    def quote_string(cls, s: str) -> str:
        quoted = shlex.quote(s)
        if quoted.startswith("'") and '"' not in quoted:
            quoted = f'"{quoted[1:-1]}"'
        return quoted


def unlink(path: os.PathLike, missing_ok: bool = False):
    try:
        os.unlink(path)
    except FileNotFoundError as exc:
        if not missing_ok:
            raise exc


def data_path(path: os.PathLike) -> Path:
    here = Path(__file__).parent
    return here / "data" / path


def deep_update(mapping: Mapping, *updating_mappings: Iterable[Mapping]) -> Mapping:
    # Brought from pydantic.utils
    # https://github.com/samuelcolvin/pydantic/blob/9d631a3429a66f30742c1a52c94ac18ec6ba848d/pydantic/utils.py#L198

    # The MIT License (MIT)
    # Copyright (c) 2017, 2018, 2019, 2020, 2021 Samuel Colvin and other contributors
    # Permission is hereby granted, free of charge, to any person obtaining a copy
    # of this software and associated documentation files (the "Software"), to deal
    # in the Software without restriction, including without limitation the rights
    # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    # copies of the Software, and to permit persons to whom the Software is
    # furnished to do so, subject to the following conditions:
    # The above copyright notice and this permission notice shall be included in all
    # copies or substantial portions of the Software.
    # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    # SOFTWARE.

    updated_mapping = mapping.copy()
    for updating_mapping in updating_mappings:
        for k, v in updating_mapping.items():
            if (
                k in updated_mapping
                and isinstance(updated_mapping[k], dict)
                and isinstance(v, dict)
            ):
                updated_mapping[k] = deep_update(updated_mapping[k], v)
            else:
                updated_mapping[k] = v
    return updated_mapping


def needs_admin(target_prefix: os.PathLike, base_prefix: os.PathLike) -> bool:
    """
    Checks if the current installation needs admin permissions.
    """
    if user_is_admin():
        return False

    if Path(target_prefix, ".nonadmin").exists():
        # This file is planted by the constructor installer
        # and signals we don't need admin permissions
        return False

    try:
        Path(target_prefix, ".nonadmin").touch()
        return False
    except Exception as exc:
        logger.debug("Attempt to write %s/.nonadmin failed.", target_prefix, exc_info=exc)

    if base_prefix == target_prefix:
        # We are already in the base env, no need to check further
        return True

    # I can't think of cases where users can't write to target_prefix but can to base
    # so maybe we can skip everything underneath?

    if Path(base_prefix, ".nonadmin").exists():
        return False

    if os.name == "nt":
        # Absence of $base_prefix/.nonadmin in Windows means we need admin permissions
        return True

    if os.name == "posix":
        # Absence of $base_prefix/.nonadmin in Linux, macOS and other posix systems
        # has no meaning for historic reasons, so let's try to see if we can
        # write to the installation root
        try:
            Path(base_prefix, ".nonadmin").touch()
        except Exception as exc:
            logger.debug("Attempt to write %s/.nonadmin failed.", target_prefix, exc_info=exc)
            return True
        else:
            return False


@lru_cache(maxsize=1)
def user_is_admin() -> bool:
    if os.name == "nt":
        from .platforms.win_utils.win_elevate import isUserAdmin

        return bool(isUserAdmin())
    elif os.name == "posix":
        # Check for root on Linux, macOS and other posix systems
        return os.getuid() == 0
    else:
        raise RuntimeError(f"Unsupported operating system: {os.name}")


def run_as_admin(argv: Sequence[str]) -> int:
    """
    Rerun this command in a new process with admin permissions.
    """
    if os.name == "nt":
        from .platforms.win_utils.win_elevate import runAsAdmin

        return runAsAdmin(argv)
    elif os.name == "posix":
        return subprocess.call(["sudo", *argv])
    else:
        raise RuntimeError(f"Unsupported operating system: {os.name}")


def python_executable(base_prefix: Optional[os.PathLike] = None) -> Sequence[str]:
    base_prefix = Path(base_prefix or DEFAULT_BASE_PREFIX)
    # menuinst might be called by a conda-standalone bundle
    # these are pyinstaller-generated, and have sys.frozen=True
    # in these cases, we prefer using the base env python to
    # avoid a 2nd decompression + hacky console, so we try that
    # first; otherwise, we use 'conda-standalone.exe python'
    if getattr(sys, "frozen", False):
        if os.name == "nt":
            base_prefix_python = base_prefix / "python.exe"
        else:
            base_prefix_python = base_prefix / "bin" / "python"
        # If the base env (installation root)
        # ships a usable Python, use that one
        if base_prefix_python.is_file():
            return (str(base_prefix_python),)
        # the base env does not have python,
        # use the conda-standalone wrapper
        return (sys.executable, "python")
    # in non-frozen executables:
    return (sys.executable,)


def elevate_as_needed(func: Callable) -> Callable:
    """
    Multiplatform decorator to run a function as a superuser, if needed.

    This depends on the presence of a `.nonadmin` file in the installation root.
    This is usually planted by the `constructor` installer if the installation
    process didn't need superuser permissions.

    In the absence of this file, we assume that we will need superuser
    permissions, so we try to run the decorated function as a superuser.
    If that fails (the user rejects the request or doesn't have permissions
    to accept it), we'll try to run it as a normal user.

    NOTE: Only functions that return None should be decorated. The function
    will run in a separate process, so we won't be able to capture the return
    value anyway.
    """

    @wraps(func)
    def wrapper_elevate(
        *args,
        target_prefix: os.PathLike = None,
        base_prefix: os.PathLike = None,
        **kwargs,
    ):
        kwargs.pop("_mode", None)
        target_prefix = target_prefix or DEFAULT_BASE_PREFIX
        base_prefix = base_prefix or DEFAULT_BASE_PREFIX
        if (
            needs_admin(target_prefix, base_prefix)
            and os.environ.get("_MENUINST_RECURSING") != "1"
        ):
            # call the wrapped func with elevated prompt...
            # from the command line; not pretty!
            try:
                if func.__module__ == "__main__":
                    import_func = (
                        f"import runpy;"
                        f"{func.__name__} = runpy.run_path('{__file__}')"
                        f"['{func.__name__}'];"
                    )
                else:
                    import_func = f"from {func.__module__} import {func.__name__};"
                env_vars = ";".join(
                    [
                        f"os.environ.setdefault('{k}', '{v}')"
                        for (k, v) in os.environ.items()
                        if k.startswith(("CONDA_", "CONSTRUCTOR_", "MENUINST_"))
                    ]
                )
                cmd = [
                    *python_executable(),
                    "-c",
                    f"import os;"
                    f"os.environ.setdefault('_MENUINST_RECURSING', '1');"
                    f"{env_vars};"
                    f"{import_func}"
                    f"{func.__name__}("
                    f"*{args!r},"
                    f"target_prefix={target_prefix!r},"
                    f"base_prefix={base_prefix!r},"
                    f"_mode='system',"
                    f"**{kwargs!r}"
                    ")",
                ]
                logger.debug("Elevating command: %s", cmd)
                return_code = run_as_admin(cmd)
            except Exception as exc:
                logger.warn("Elevation failed! Falling back to user mode.", exc_info=exc)
            else:
                os.environ.pop("_MENUINST_RECURSING", None)
                if return_code == 0:  # success, we are done
                    return
        elif user_is_admin():
            # We are already running as admin, no need to elevate
            return func(
                target_prefix=target_prefix,
                base_prefix=base_prefix,
                _mode="system",
                *args,
                **kwargs,
            )
        # We have not returned yet? Well, let's try as a normal user
        return func(
            target_prefix=target_prefix,
            base_prefix=base_prefix,
            _mode="user",
            *args,
            **kwargs,
        )

    return wrapper_elevate


def _test_elevation(
    target_prefix: Optional[os.PathLike] = None,
    base_prefix: Optional[os.PathLike] = None,
    _mode: _UserOrSystem = "user",
):
    if os.name == "nt":
        if base_prefix:
            output = os.path.join(base_prefix, "_test_output.txt")
        else:
            output = "_test_output.txt"
        out = open(output, "a")
    else:
        out = sys.stdout
    print(
        "user_is_admin():",
        user_is_admin(),
        "env_var:",
        os.environ.get("MENUINST_TEST", "N/A"),
        "_mode:",
        _mode,
        file=out,
    )
    if os.name == "nt":
        out.close()


def logged_run(args, check=False, log=True, **kwargs) -> subprocess.CompletedProcess:
    process = subprocess.run(args, capture_output=True, text=True, **kwargs)
    if log:
        logger.debug("%s returned %d", process.args, process.returncode)
        logger.debug("stdout:\n%s", process.stdout)
        logger.debug("stderr:\n%s", process.stderr)
    if check:
        process.check_returncode()
    return process



import sys
from subprocess import check_output

import pytest

cwp = pytest.importorskip("menuinst._legacy.cwp", reason="Windows only")


def test_cwp():
    out = check_output(
        [
            sys.executable,
            cwp.__file__,
            sys.prefix,
            "python",
            "-c",
            "import sys; print(sys.prefix)",
        ],
        text=True,
    )
    assert out.strip() == sys.prefix
